/******************************************************************************************
 fillout_PDF_forms.c - Reads results output files and then scans through a 
	corresponding pre-marked-up PDF file searching for unique string(s) for
	each entry and performs the designated replacement when found.  
	Then it saves the file back out to a new name.

	All replaced phrases must have the exact same length.  (char-by-char replacements)
	So file size does not (and cannot) change.

	For example:
	
	   src/formdata/form_meta.data:

		NJ_1040_2013 1 8 99997	 src/formdata/form_nj_1040.pdf
		form_1040  2   9 {Q6  3 {Q7  src/formdata/form_1040.pdf
			   ^   ^  ^  
			   |   |  \ Unique marker chars
			   |   \ Tells length of replacement string
			   \ Type (2 means dollars and cents)

		Replacement_Type:
			1: Display only whole integer dollar amounts (no cents).
			2. Display dollars and cents in separate pdf fields.
			3. Display dollars.cents in single field.

		ForceCents - For replacement-type 2, in the case where 
			the cents part marker did not directly precede the dollar-part
			marker, this option forces searching through the remaineder of
			the pdf file for the next cents marker.

		no_zero_entries - Does not make entries on lines with 0.00 amounts.

	   form_1040_out.txt:		(text)
			L7 = 48456.23
			L8a =  67.21
			L9 = 253.91

	   src/formdata/form_1040.pdf:	(binary)
			... {Q6L7_000000
			... {Q6L8a_00000
			... {Q6L9_000000

 But for rapid development, we could just compile this knowledge into the program.
 Then as a later step, we'll extract this as a metadata file.

 Run:
	fillout_pdf_forms  form_meta.data  return_out.txt  premarkedup_form.pdf

 Compile:
	cc -g fillout_PDF_forms.c -o fillout_PDF_forms
*******************************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <sys/stat.h>

int verbose=0;
float thisversion=1.02;

#define MAXLINE 10000

char line[MAXLINE], word1[MAXLINE], word2[MAXLINE], origline[MAXLINE];
char results_file[MAXLINE], *outputname=0;
int replacement_type=1;
char *srchstrng1=0, *srchstrng2=0, rplcstr1[1000], rplcstr2[1000];
int srch1_len, srch2_len=0, force_cents=0, no_zero_entries=0;
char *form_is=0;
int detected_8949short=0, detected_8949long=0, detected_Sched_D=0, cg_list1=1, cg_list2=1;
int form8949_bnr=0;

struct nvpair
 {
  char *label, *value;
  int used;
  struct nvpair *nxt;
 } *results_list=0, *alias_list=0, *symb_list=0, *special_list=0;


void next_word( char *line, char *word, char *delim )
{
 int k=0, j=0, m=0, flag=1;
 while ((line[k] !='\0') && (flag))	 /* Eat away preceding garbage */
  {
   j = 0;
   while ((delim[j] != '\0') && (line[k] != delim[j])) j = j + 1;
   if (line[k] == delim[j]) k++;
   else  flag = 0;
  }
 while ((line[k] != '\0') && (!flag))
  {
   word[m++] = line[k++];
   if (line[k] != '\0')
    {
     j = 0;
     while ((delim[j] != '\0') && (line[k] != delim[j])) j = j + 1;
     if (line[k] == delim[j]) flag = 1;
    }
  }
 j = 0;		 	/* Shorten line. */
 while (line[k] != '\0') { line[j++] = line[k++]; }
 line[j] = '\0';	/* Terminate the char-strings. */
 word[m] = '\0';
}


void last_word( char *line, char *word, char *delim )	/* Like next_word, but works from back of string. */
{
 int j, k, flag=1;
 /* Consume any ending white-space. */
 j = strlen( line ) - 1;
 if (j < 0) { word[0] = '\0';  return; }
 while ((j >= 0) && (flag))
  {
   k = 0;
   while ((delim[k] != '\0') && (line[j] != delim[k])) k++;
   if (delim[k] == '\0')  flag = 0;  else  j--; 
  }
 if (j < 0) { line[0] = '\0';  word[0] = '\0';  return; }
 line[j+1] = '\0';
 j--;
 while ((j >= 0) && (!flag))
  {
   k = 0;
   while ((delim[k] != '\0') && (line[j] != delim[k])) k++;
   if (delim[k] == '\0')  j--;  else  flag = 1;
  }
 j++;
 k = 0;
 do { word[k] = line[j+k];  k++; } while (word[k-1] != '\0');
 /* Shorten line. */
 line[j] = '\0';
}
 

struct nvpair *new_item( char *label, char *value )
{
 struct nvpair *item;
 item = (struct nvpair *)malloc( sizeof( struct nvpair ) );
 item->label = strdup( label );
 item->value = strdup( value );
 item->used = 0;
 return item;
}


void add_entry( char *label, char *value )	/* Builds list of result items (tax-return line answer values). */
{						/* Example:  name="L7" value="23,456.00"  */
 struct nvpair *item;
 if ((no_zero_entries) && (strcmp( value, "0.00" ) == 0)) return;
 item = new_item( label, value );
 item->nxt = results_list;
 results_list = item;
}


void add_alias( char *label, char *value )	/* Builds list of line-labels and corresponding marker-symbols. */
{						/* Example:  name="L7" value="{Q6109"	*/
 struct nvpair *item;
 item = new_item( label, value );
 item->nxt = alias_list;
 alias_list = item;
}


void add_symb( char *symb, char *value )	/* Builds list of marker-symbols and their corresponding replacement values. */
{						/* Example:  symb="{Q90170000001" value="Bank 1"  */
 struct nvpair *item;
 item = new_item( symb, value );
 if (verbose) printf("Adding Symbol '%s' = '%s'\n", symb, value );
 item->nxt = symb_list;
 symb_list = item;
}


void add_special_rule( char *label, char *rule ) /* Builds list of special rules for lines that are to be handled differently. */
{						 /* Example:  Special_Rule  L1a_A:  full-line */
 struct nvpair *item;
 item = new_item( label, rule );
 if (verbose) printf("Adding Special Rule '%s' = '%s'\n", label, rule );
 item->nxt = special_list;
 special_list = item;
}


int special_line( char *label )
{
 struct nvpair *item;
 item = special_list;
 while (item != 0) { if (strcmp(item->label, label) == 0) return 1; else item = item->nxt; }
 return 0;
}


void cleanline( char *line )	/* Clean any any weird characters like <cr>/<lf> from string. */
{
 int j=0;
 while (line[j] != '\0') 
  {
   if (line[j] == '\n') line[j] = '\0';
   if (line[j] == '\r') line[j] = '\0';
   j++;
  }
}


void Clear_Sched_D_Entries()
{
 add_symb( "{Q9D1bd_0000", "" );
 add_symb( "{Q9D1be_0000", "" );
 add_symb( "{Q9D1bh_0000", "" );
 add_symb( "{Q9D2d_00000", "" );
 add_symb( "{Q9D2e_00000", "" );
 add_symb( "{Q9D2h_00000", "" );
 add_symb( "{Q9D3d_00000", "" );
 add_symb( "{Q9D3e_00000", "" );
 add_symb( "{Q9D3h_00000", "" );
 add_symb( "{Q9D8bd_0000", "" );
 add_symb( "{Q9D8be_0000", "" );
 add_symb( "{Q9D8bh_0000", "" );
 add_symb( "{Q9D9d_00000", "" );
 add_symb( "{Q9D9e_00000", "" );
 add_symb( "{Q9D9h_00000", "" );
 add_symb( "{Q9D10d_0000", "" );
 add_symb( "{Q9D10e_0000", "" );
 add_symb( "{Q9D10h_0000", "" );
}


void Clear_8949_Entries( int ln1, int ln2 )
{ /* Set any remaining form-8949 entries to blank to clear them. */
 int lnum;
 char symb[100];
 lnum = ln1;
 while (lnum <= 14)
  {
   if (lnum < 10) sprintf(symb,"{Q91h_0%d_0000", lnum );  else sprintf(symb,"{Q91h_%d_0000", lnum );
   add_symb( symb, "" );
   if (lnum < 10) sprintf(symb,"{Q91e_0%d_0000", lnum );  else sprintf(symb,"{Q91e_%d_0000", lnum );
   add_symb( symb, "" );
   if (lnum < 10) sprintf(symb,"{Q91d_0%d_0000", lnum );  else sprintf(symb,"{Q91d_%d_0000", lnum );
   add_symb( symb, "" );
   if (lnum < 10) sprintf(symb,"{Q91c_0%d_0", lnum );  else sprintf(symb,"{Q91c_%d_0", lnum );
   add_symb( symb, "" );
   if (lnum < 10) sprintf(symb,"{Q91b_0%d_0", lnum );  else sprintf(symb,"{Q91b_%d_0", lnum );
   add_symb( symb, "" );
   if (lnum < 10) sprintf(symb,"{Q91a_0%d_01234567890123456789", lnum );  else sprintf(symb,"{Q91a_%d_01234567890123456789", lnum );
   add_symb( symb, "" );
   lnum++;
  }
 lnum = ln2;
 while (lnum <= 14)
  {
   if (lnum < 10) sprintf(symb,"{Q92h_0%d_0000", lnum );  else sprintf(symb,"{Q92h_%d_0000", lnum );
   add_symb( symb, "" );
   if (lnum < 10) sprintf(symb,"{Q92e_0%d_0000", lnum );  else sprintf(symb,"{Q92e_%d_0000", lnum );
   add_symb( symb, "" );
   if (lnum < 10) sprintf(symb,"{Q92d_0%d_0000", lnum );  else sprintf(symb,"{Q92d_%d_0000", lnum );
   add_symb( symb, "" );
   if (lnum < 10) sprintf(symb,"{Q92c_0%d_0", lnum );  else sprintf(symb,"{Q92c_%d_0", lnum );
   add_symb( symb, "" );
   if (lnum < 10) sprintf(symb,"{Q92b_0%d_0", lnum );  else sprintf(symb,"{Q92b_%d_0", lnum );
   add_symb( symb, "" );
   if (lnum < 10) sprintf(symb,"{Q92a_0%d_01234567890123456789", lnum );  else sprintf(symb,"{Q92a_%d_01234567890123456789", lnum );
   add_symb( symb, "" );
   lnum++;
  }
 if (!detected_8949short)
  {
   add_symb( "{Q91d_00000000", "" );
   add_symb( "{Q91e_00000000", "" );
   add_symb( "{Q91h_00000000", "" );
  }
 if (!detected_8949long)
  {
   add_symb( "{Q92d_00000000", "" );
   add_symb( "{Q92e_00000000", "" );
   add_symb( "{Q92h_00000000", "" );
  }
}



void handle_scheduleD_entries( char *word1, char *line )
{
 char symb[100]="{Q9", value[200], prefix[20]="{Q9";
 
 word1[strlen(word1)-1] = '\0';	/* Chop colon off end of line lable. (ex. "D1b:" becomes "D1b".) */
 strcat( prefix, word1 );

 strcpy( symb, prefix );
 strcat( symb, "d_");
 while (strlen(symb) < 12) strcat(symb, "0");
 next_word( line, value, " \t=\n\t\r" );
 add_symb( symb, value );

 next_word( line, value, " \t=\n\t\r" );	/* Should be 'e'. */
 strcpy( symb, prefix );
 strcat( symb, "e_");
 while (strlen(symb) < 12) strcat(symb, "0");
 next_word( line, value, " \t=\n\t\r" );
 add_symb( symb, value );

 next_word( line, value, " \t=\n\t\r" );	/* Should be 'h'. */
 strcpy( symb, prefix );
 strcat( symb, "h_");
 while (strlen(symb) < 12) strcat(symb, "0");
 next_word( line, value, " \t=\n\t\r" );
 add_symb( symb, value );
 detected_Sched_D = 1;
}


void handle_form8949_entries_shortterm( FILE *infile, int record )
{
 int lnum=0;
 char symb[100];

 detected_8949short = 1;
 do	/* Get Part-I. */
  {fgets( line, MAXLINE, infile );  if (feof(infile)) { printf("Error1: Premature end of file.\n");  exit(1); } }
 while (strstr(line,"---------------------") == 0);
 fgets( line, MAXLINE, infile );	/* Get Part-I. */
 while (strstr(line,"---------------------") == 0)
  {
   lnum++;
   last_word( line, word1, " \t" );
   if (lnum < 10) sprintf(symb,"{Q91h_0%d_0000", lnum );  else sprintf(symb,"{Q91h_%d_0000", lnum );
   if (record) add_symb( symb, word1 );

   last_word( line, word1, " \t" );
   if (lnum < 10) sprintf(symb,"{Q91e_0%d_0000", lnum );  else sprintf(symb,"{Q91e_%d_0000", lnum );
   if (record) add_symb( symb, word1 );

   last_word( line, word1, " \t" );
   if (lnum < 10) sprintf(symb,"{Q91d_0%d_0000", lnum );  else sprintf(symb,"{Q91d_%d_0000", lnum );
   if (record) add_symb( symb, word1 );

   last_word( line, word1, " \t" );
   if (lnum < 10) sprintf(symb,"{Q91c_0%d_0", lnum );  else sprintf(symb,"{Q91c_%d_0", lnum );
   if (record) add_symb( symb, word1 );

   last_word( line, word1, " \t" );
   if (lnum < 10) sprintf(symb,"{Q91b_0%d_0", lnum );  else sprintf(symb,"{Q91b_%d_0", lnum );
   if (record) add_symb( symb, word1 );

   if (lnum < 10) sprintf(symb,"{Q91a_0%d_01234567890123456789", lnum );  else sprintf(symb,"{Q91a_%d_01234567890123456789", lnum );
   line[32] = '\0';
   if (record) add_symb( symb, line );
   fgets( line, MAXLINE, infile );	/* Get Part-I. */
  }
 if (lnum > 14)
  printf("Too many cap-gains entries for fill-in program as presently configured.\nShort-term entries 14-%d not shown.", lnum );
 lnum++;
 if (record) cg_list1 = lnum;
 fgets( line, MAXLINE, infile ); 	/* Get "Totals" */
 next_word( line, word1, " \t" );
 next_word( line, word1, " \t" );
 if (strcmp(word1,"Totals:") != 0) printf("Warning: CapGains expected 'Totals:' but found '%s'.\n", word1 );
 next_word( line, word1, " \t" );
 if (record) add_symb( "{Q91d_00000000", word1 );
 next_word( line, word1, " \t" );
 if (record) add_symb( "{Q91e_00000000", word1 );
 next_word( line, word1, " \t" );
 if (record) add_symb( "{Q91h_00000000", word1 );
}

void handle_form8949_entries_longterm( FILE *infile, int record )
{
 int lnum=0;
 char symb[100];

 detected_8949long = 1;
 do	/* Get Part-II. */
  {fgets( line, MAXLINE, infile );  if (feof(infile)) { printf("Error2: Premature end of file.\n");  exit(1); } }
 while (strstr(line,"---------------------") == 0);
 fgets( line, MAXLINE, infile );	/* Get Part-II. */
 while (strstr(line,"---------------------") == 0)
  {
   lnum++;
   last_word( line, word1, " \t" );
   if (lnum < 10) sprintf(symb,"{Q92h_0%d_0000", lnum );  else sprintf(symb,"{Q92h_%d_0000", lnum );
   if (record) add_symb( symb, word1 );

   last_word( line, word1, " \t" );
   if (lnum < 10) sprintf(symb,"{Q92e_0%d_0000", lnum );  else sprintf(symb,"{Q92e_%d_0000", lnum );
   if (record) add_symb( symb, word1 );

   last_word( line, word1, " \t" );
   if (lnum < 10) sprintf(symb,"{Q92d_0%d_0000", lnum );  else sprintf(symb,"{Q92d_%d_0000", lnum );
   if (record) add_symb( symb, word1 );

   last_word( line, word1, " \t" );
   if (lnum < 10) sprintf(symb,"{Q92c_0%d_0", lnum );  else sprintf(symb,"{Q92c_%d_0", lnum );
   if (record) add_symb( symb, word1 );

   last_word( line, word1, " \t" );
   if (lnum < 10) sprintf(symb,"{Q92b_0%d_0", lnum );  else sprintf(symb,"{Q92b_%d_0", lnum );
   if (record) add_symb( symb, word1 );

   if (lnum < 10) sprintf(symb,"{Q92a_0%d_01234567890123456789", lnum );  else sprintf(symb,"{Q92a_%d_01234567890123456789", lnum );
   line[32] = '\0';
   if (record) add_symb( symb, line );
   fgets( line, MAXLINE, infile );	/* Get Part-I. */
  }
 if (lnum > 14)
  printf("Too many cap-gains entries for fill-in program as presently configured.\nLong-term entries 14-%d not shown.", lnum );
 lnum++;
 if (record) cg_list2 = lnum;
 fgets( line, MAXLINE, infile ); 	/* Get "Totals" */
 next_word( line, word1, " \t" );
 next_word( line, word1, " \t" );
 if (strcmp(word1,"Totals:") != 0) printf("Warning: CapGains expected 'Totals:' but found '%s'.\n", word1 );
 next_word( line, word1, " \t" );
 if (record) add_symb( "{Q92d_00000000", word1 );
 next_word( line, word1, " \t" );
 if (record) add_symb( "{Q92e_00000000", word1 );
 next_word( line, word1, " \t" );
 if (record) add_symb( "{Q92h_00000000", word1 );
}


void handle_form8582_entries( FILE *infile )
{
 int lnum=0;
 char symb[100];

 do	/* Get Worksheet-1. */
  {fgets( line, MAXLINE, infile );  if (feof(infile)) { printf("Error3: Premature end of file.\n");  exit(1); } }
 while (strstr(line,"---------------------") == 0);
 do {fgets( line, MAXLINE, infile );  if (feof(infile)) { printf("Error4: Premature end of file.\n");  exit(1); } }
 while (strstr(line,"---------------------") == 0);
 fgets( line, MAXLINE, infile );
 while (strstr(line,"---------------------") == 0)
  {
   last_word( line, word1, " \t" );
   if (word1[0] == '-')
    {
     sprintf(symb,"{Q9WS1e%d_012345", lnum );
     add_symb( symb, word1 );
     sprintf(symb,"{Q9WS1d%d_012345", lnum );
     add_symb( symb, "" );
    }
   else
    {
     sprintf(symb,"{Q9WS1e%d_012345", lnum );
     add_symb( symb, "" );
     sprintf(symb,"{Q9WS1d%d_012345", lnum );
     add_symb( symb, word1 );
    }

   last_word( line, word1, " \t" );
   sprintf(symb,"{Q9WS1c%d_012345", lnum );
   add_symb( symb, word1 );

   last_word( line, word1, " \t" );
   sprintf(symb,"{Q9WS1b%d_012345", lnum );
   add_symb( symb, word1 );

   last_word( line, word1, " \t" );
   sprintf(symb,"{Q9WS1a%d_012345", lnum );
   add_symb( symb, word1 );

   sprintf(symb,"{Q9WS1_%d_01234567890123456789", lnum );
   line[28] = '\0';
   add_symb( symb, line );
   fgets( line, MAXLINE, infile );
   lnum++;
  }
 /* Set any remaining ones to blank to clear them. */
 while (lnum <= 4)
  {
   sprintf(symb,"{Q9WS1_%d_01234567890123456789", lnum );  add_symb( symb, "" );
   sprintf(symb,"{Q9WS1a%d_012345", lnum );   		   add_symb( symb, "" );
   sprintf(symb,"{Q9WS1b%d_012345", lnum );   		   add_symb( symb, "" );
   sprintf(symb,"{Q9WS1c%d_012345", lnum );   		   add_symb( symb, "" );
   sprintf(symb,"{Q9WS1d%d_012345", lnum );   		   add_symb( symb, "" );
   sprintf(symb,"{Q9WS1e%d_012345", lnum );   		   add_symb( symb, "" );
   lnum++;
  }
 fgets( line, MAXLINE, infile ); 	/* Get "Totals" */
 next_word( line, word1, " \t" );
 add_symb( "{Q9WS1a_0123456", word1 );
 next_word( line, word1, " \t" );
 add_symb( "{Q9WS1b_0123456", word1 );
 next_word( line, word1, " \t" );
 add_symb( "{Q9WS1c_0123456", word1 );

 lnum = 0;
 do	/* Get Worksheet-2. */
  {fgets( line, MAXLINE, infile );  if (feof(infile)) { printf("Error5: Premature end of file.\n");  exit(1); } }
 while (strstr(line,"---------------------") == 0);
 do {fgets( line, MAXLINE, infile );  if (feof(infile)) { printf("Error6: Premature end of file.\n");  exit(1); } }
 while (strstr(line,"---------------------") == 0);
 fgets( line, MAXLINE, infile );
 while (strstr(line,"---------------------") == 0)
  {
   last_word( line, word1, " \t" );
   sprintf(symb,"{Q9WS2c%d_012345", lnum );
   add_symb( symb, word1 );

   last_word( line, word1, " \t" );
   sprintf(symb,"{Q9WS2b%d_012345", lnum );
   add_symb( symb, word1 );

   last_word( line, word1, " \t" );
   sprintf(symb,"{Q9WS2a%d_012345", lnum );
   add_symb( symb, word1 );

   sprintf(symb,"{Q9WS2_%d_01234567890123456789", lnum );
   line[28] = '\0';
   add_symb( symb, line );
   fgets( line, MAXLINE, infile );
   lnum++;
  }
 /* Set any remaining ones to blank to clear them. */
 while (lnum <= 4)
  {
   sprintf(symb,"{Q9WS2_%d_01234567890123456789", lnum );  add_symb( symb, "" );
   sprintf(symb,"{Q9WS2a%d_012345", lnum );   		   add_symb( symb, "" );
   sprintf(symb,"{Q9WS2b%d_012345", lnum );   		   add_symb( symb, "" );
   sprintf(symb,"{Q9WS2c%d_012345", lnum );   		   add_symb( symb, "" );
   lnum++;
  }
 fgets( line, MAXLINE, infile ); 	/* Get "Totals" */
 next_word( line, word1, " \t" );
 add_symb( "{Q9WS2a_0123456", word1 );
 next_word( line, word1, " \t" );
 add_symb( "{Q9WS2b_0123456", word1 );

 lnum = 0;
 do	/* Get Worksheet-3. */
  {fgets( line, MAXLINE, infile );  if (feof(infile)) { printf("Error7: Premature end of file.\n");  exit(1); } }
 while (strstr(line,"---------------------") == 0);
 do {fgets( line, MAXLINE, infile );  if (feof(infile)) { printf("Error8: Premature end of file.\n");  exit(1); } }
 while (strstr(line,"---------------------") == 0);
 fgets( line, MAXLINE, infile );
 while (strstr(line,"---------------------") == 0)
  {
   last_word( line, word1, " \t" );
   if (word1[0] == '-')
    {
     sprintf(symb,"{Q9WS3e%d_012345", lnum );
     add_symb( symb, word1 );
     sprintf(symb,"{Q9WS3d%d_012345", lnum );
     add_symb( symb, "" );
    }
   else
    {
     sprintf(symb,"{Q9WS3e%d_012345", lnum );
     add_symb( symb, "" );
     sprintf(symb,"{Q9WS3d%d_012345", lnum );
     add_symb( symb, word1 );
    }

   last_word( line, word1, " \t" );
   sprintf(symb,"{Q9WS3c%d_012345", lnum );
   add_symb( symb, word1 );

   last_word( line, word1, " \t" );
   sprintf(symb,"{Q9WS3b%d_012345", lnum );
   add_symb( symb, word1 );

   last_word( line, word1, " \t" );
   sprintf(symb,"{Q9WS3a%d_012345", lnum );
   add_symb( symb, word1 );

   sprintf(symb,"{Q9WS3_%d_01234567890123456789", lnum );
   line[28] = '\0';
   add_symb( symb, line );
   fgets( line, MAXLINE, infile );
   lnum++;
  }
 /* Set any remaining ones to blank to clear them. */
 while (lnum <= 4)
  {
   sprintf(symb,"{Q9WS3_%d_01234567890123456789", lnum );  add_symb( symb, "" );
   sprintf(symb,"{Q9WS3a%d_012345", lnum );   		   add_symb( symb, "" );
   sprintf(symb,"{Q9WS3b%d_012345", lnum );   		   add_symb( symb, "" );
   sprintf(symb,"{Q9WS3c%d_012345", lnum );   		   add_symb( symb, "" );
   sprintf(symb,"{Q9WS3d%d_012345", lnum );   		   add_symb( symb, "" );
   sprintf(symb,"{Q9WS3e%d_012345", lnum );   		   add_symb( symb, "" );
   lnum++;
  }
 fgets( line, MAXLINE, infile ); 	/* Get "Totals" */
 next_word( line, word1, " \t" );
 add_symb( "{Q9WS3a_0123456", word1 );
 next_word( line, word1, " \t" );
 add_symb( "{Q9WS3b_0123456", word1 );
 next_word( line, word1, " \t" );
 add_symb( "{Q9WS3c_0123456", word1 );

 for (lnum = 0; lnum <= 3; lnum++)
  {
   sprintf(symb,"{Q9WS4_%d_01234567890123456789", lnum );  add_symb( symb, "" );
   sprintf(symb,"{Q9WS4a%d_012345", lnum );   		   add_symb( symb, "" );
   sprintf(symb,"{Q9WS4b%d_012345", lnum );   		   add_symb( symb, "" );
   sprintf(symb,"{Q9WS4c%d_012345", lnum );   		   add_symb( symb, "" );
   sprintf(symb,"{Q9WS4d%d_012345", lnum );   		   add_symb( symb, "" );

   sprintf(symb,"{Q9WS5_%d_01234567890123456789", lnum );  add_symb( symb, "" );
   sprintf(symb,"{Q9WS5a%d_012345", lnum );   		   add_symb( symb, "" );
   sprintf(symb,"{Q9WS5b%d_012345", lnum );   		   add_symb( symb, "" );
   sprintf(symb,"{Q9WS5c%d_012345", lnum );   		   add_symb( symb, "" );
  }

} /*handle_form8582_entries*/



void read_in_results_file( char *fname )	/* Reads the answers in from "..._out.txt" files. */
{
 FILE *infile;
 if (verbose) printf("Reading results file.\n");
 infile = fopen( fname, "rb" );
 if (infile == 0) { printf("Error: Cannot read '%s'\n", fname );  exit(1); }
 fgets( line, MAXLINE, infile );
 while (!feof(infile))
  {
   strcpy( origline, line );
   next_word( line, word1, " \t=\n\t\r" );
   next_word( line, word2, " \t=\n\r" );
   if (word2[0] != '\0')
    {
     if ((word1[0] == 'D') && (strstr(word1,":") != 0) && (strcmp(word2,"d") == 0))
      handle_scheduleD_entries( word1, line );
     else
     if (strstr(origline,"Form 8949 Part-I") == &(origline[0]))
      { int record;
       if ((form8949_bnr == 0) && (strstr(origline,"(A) Basis Reported") != 0)) record = 1;
       else
       if ((form8949_bnr != 0) && (strstr(origline,"(A) Basis Reported") == 0)) record = 1;
       else
	record = 0;
       if (strstr(origline,"Short-Term Cap Gains+Losses,") != 0)
	handle_form8949_entries_shortterm( infile, record );
       else
	handle_form8949_entries_longterm( infile, record );
      }
     else
     if ((strcmp(word1,"---") == 0) && (strcmp(word2,"---") == 0) && (strstr(line, "Worksheet 1 --- ") != 0))
      handle_form8582_entries( infile );
     else
     if ((strcmp(word1,"AMT") == 0) && (strcmp(word2,"Form") == 0) && (strstr(line," 6251, ") != 0))
      {
	next_word( line, word1, " \t=\n\t\r" );	 /* gets '6251' */
	next_word( line, word2, " \t=\n\t\r" );
	strcpy( word1, "AMT_Form_6251_" );
	strcat( word1, word2 );
	next_word( line, word2, " \t=\n\t\r" );
	add_entry( word1, word2 );
      }
     else
     if (special_line( word1 ))
      {
       strcat( word2, " " );
       cleanline( line );
       strcat( word2, line );
       add_entry( word1, word2 );
      }
     else
      add_entry( word1, word2 );
    }
   fgets( line, MAXLINE, infile );
  }
 fclose(infile);
 if (verbose) printf("Done reading results file.\n\n");
}


void read_metadata( char *fname )
{
 FILE *infile;
 infile = fopen( fname, "rb" );
 if (infile == 0) { printf("Error: Cannot read '%s'\n", fname );  exit(1); }

 fscanf(infile,"%s",word1);
 if (strcmp( word1, "Replacement_Type" ) != 0) 
  { printf("Error: reading metadata file '%s' on 'Replacement_Type', got '%s' instead.\n", fname, word1 );  exit(1); }
 if (fscanf(infile,"%d", &replacement_type ) != 1)
  { printf("Error Bad value in metadata file '%s' on 'Replacement_Type'.\n", fname );  exit(1); }

 fscanf(infile,"%s",word1);
 if (strcmp( word1, "SrchStrng1" ) != 0) 
  { printf("Error: reading metadata file '%s' on 'SrchStrng1', got '%s' instead.\n", fname, word1 );  exit(1); }
 fscanf(infile,"%s",word1);
 srchstrng1 = strdup( word1 );

 fscanf(infile,"%s",word1);
 if (strcmp( word1, "Srch1_Len" ) != 0) 
  { printf("Error: reading metadata file '%s' on 'Srch1_Len', got '%s' instead.\n", fname, word1 );  exit(1); }
 if (fscanf(infile,"%d", &srch1_len ) != 1)
  { printf("Error Bad value in metadata file '%s' on 'Srch1_Len'.\n", fname );  exit(1); }

 fscanf(infile,"%s",word1);
 if (strcmp( word1, "SrchStrng2" ) != 0) 
  { printf("Error: reading metadata file '%s' on 'SrchStrng2', got '%s' instead.\n", fname, word1 );  exit(1); }
 fscanf(infile,"%s",word1);
 srchstrng2 = strdup( word1 );

 fscanf(infile,"%s",word1);
 if (strcmp( word1, "Srch2_Len" ) != 0) 
  { printf("Error: reading metadata file '%s' on 'Srch2_Len', got '%s' instead.\n", fname, word1 );  exit(1); }
 if (fscanf(infile,"%d", &srch2_len ) != 1)
  { printf("Error Bad value in metadata file '%s' on 'Srch2_Len'.\n", fname );  exit(1); }

 fscanf(infile,"%s",word1);
 while (!feof(infile))
  {
   if (strcmp( word1, "Alias" ) == 0) 
    {
     fscanf(infile,"%s",word1);
     fscanf(infile,"%s",word2);
     add_alias( word1, word2 );
    }
   else
   if (strcmp( word1, "Special_Rule" ) == 0) 
    {
     fscanf(infile,"%s",word1);
     fscanf(infile,"%s",word2);
     add_special_rule( word1, word2 );
    }
   else
   if (strcmp( word1, "ForceCents") == 0)
    force_cents = 1;	/* If cents marker did not precede a given dollars marker, force search through file to find next cents marker. */
   else
   if (strcmp( word1, "no_zero_entries") == 0)
    no_zero_entries = 1;
   else
   if (strcmp( word1, "CleanUps" ) == 0) 
    {
     fscanf(infile,"%s",word1);
     if (strcmp( word1, "Form_8949" ) == 0) form_is = strdup( word1 );
     else
     if (strcmp( word1, "Sched_D" ) == 0) form_is = strdup( word1 );
     else 
      printf("Error: Unknown form '%s'\n", word1 );
    }
   else
    { printf("Error: Unknown primitive '%s' in '%s'\n", word1, fname );  exit(1); }
   fscanf(infile,"%s",word1);
  }
 if (verbose) printf("Done reading metadata.\n\n");
 fclose(infile);
}


void pull_comment( char *line, char *word )
{
 int j=0, k=0;
 while ((line[j] != '\0') && (line[j] != '{')) j++;
 if (line[j] != '\0')
  {
   j++;
   while ((line[j+k] != '\0') && (line[j+k] != '}'))
    {
     word[k] = line[j+k];  k++;
    }
  }
 word[k] = '\0';
}


void read_tax_input_for_schedule_data( char *fname )	/* Read input files to get certain data (Sched-B Int+Divs), */
{							/* since those data does not get repeated in "...out.txt" files. */
 int state=0, k1=0, k2=0;
 char symb[100]; 
 FILE *infile;

 infile = fopen( fname, "rb" );
 if (infile == 0) { printf("Could not read file '%s'\n", fname );  exit(1); }
 fgets( line, 200, infile );
 while (!feof(infile))
  {
   next_word( line, word1, " \t\n\r" );
   switch (state)
    {
     case 0:
	if (strcmp( word1, "L8a" ) == 0) state = 8;  else
	if (strcmp( word1, "L9" ) == 0) state = 9;  
	break;
     case 8:
	if (word1[0] == ';')
	 state = 0;
	else
	if (word1[0] != '\0')
	 {
	  pull_comment( line, word2 );
	  sprintf(symb,"{Q9B001%c", k1 + 97 );
	  strcat( symb, "_01234567890123456789012345678901234567890123456789" );
	  add_symb( symb, word2 );
	  sprintf(symb,"{Q9B01%c", k1 + 97 );
	  strcat( symb, "_0000" );
	  add_symb( symb, word1 );
	  k1++;
     	 }
	break;
     case 9:
	if (word1[0] == ';')
	 state = 0;
	else
	if (word1[0] != '\0')
	 {
	  pull_comment( line, word2 );
	  sprintf(symb,"{Q9B005%c", k2 + 97 );
	  strcat( symb, "_01234567890123456789012345678901234567890" );
	  add_symb( symb, word2 );
	  sprintf(symb,"{Q9B05%c", k2 + 97 );
	  strcat( symb, "_0000" );
	  add_symb( symb, word1 );
	  k2++;
     	 }
	break;
    }
   fgets( line, 200, infile );
  }
 fclose(infile);
 /* Create blank symbols for ones not used, so they remove markers in pdf file. */
 while (k1 <= 13)
  {
   sprintf(symb,"{Q9B001%c", k1 + 97 );
   strcat( symb, "_01234567890123456789012345678901234567890123456789" );
   add_symb( symb, "" );
   sprintf(symb,"{Q9B01%c", k1 + 97 );
   strcat( symb, "_0000" );
   add_symb( symb, "" );
   k1++;
  }
 while (k2 <= 15)
  {
   sprintf(symb,"{Q9B005%c", k2 + 97 );
   strcat( symb, "_01234567890123456789012345678901234567890" );
   add_symb( symb, "" );
   sprintf(symb,"{Q9B05%c", k2 + 97 );
   strcat( symb, "_0000" );
   add_symb( symb, "" );
   k2++;
  }
}


char *fbuf;
int filesz;


int chkstr( char *fbuf, int j, char *srchstr )
{
 int k=0;
 while ((j+k < filesz) && (fbuf[j+k] == srchstr[k])) k++;
 if (srchstr[k] == '\0')
  return 1;
 else
  return 0;
}



void replace_str( char *fbuf, int j, char *rplcstr )
{
 int k=0;
 while ((j+k < filesz) && (rplcstr[k] != '\0')) 
  {
    fbuf[j+k] = rplcstr[k];
    k++;
  }
}


void set_with_whitespaces( char *wrd, int len )
{ int j;
  for (j=0; j < len; j++) wrd[j] = ' ';
  wrd[len] = '\0';
}


void prepad_with_whitespace( char *wrd, int len )
{ int j, k;
  char newword[1000]="";
  j = strlen( wrd );
  k = 0;
  while (k < len - j) { newword[k++] = ' '; }
  newword[k] = '\0';
  strcat( newword, wrd );
  strcpy( wrd, newword );
}


int m_round( double x )
{
 if (x >= 0.0) return (int)(x + 0.5);
 else	       return  (int)(x - 0.5);
}


void prepare_replacement_strings( struct nvpair *item, char *rplcstr1, int srch1_len, char *rplcstr2, int srch2_len )
{
 double value;

 if ((replacement_type != 3) && (sscanf(item->value,"%lf", &value) == 1))
  { /*isnumeric*/
   if (replacement_type == 1)
    {
     sprintf(rplcstr1,"%d", m_round( value ) );
     prepad_with_whitespace( rplcstr1, srch1_len );
    }
   else
    {
     sprintf(rplcstr1,"%d", (int)value );
     prepad_with_whitespace( rplcstr1, srch1_len );
     sprintf(rplcstr2,"%3d", (int)(100.0 * (value - (int)value + 0.005)) );
     if (rplcstr2[1] == ' ') rplcstr2[1] = '0';
    }
  } /*isnumeric*/
 else
  { /*alpha*/
    strcpy( rplcstr1, item->value );
    rplcstr1[ srch1_len ] = '\0';
    prepad_with_whitespace( rplcstr1, srch1_len );
    set_with_whitespaces( rplcstr2, 3 );
  } /*alpha*/

 item->used = item->used + 1;
}


void replace_str_direct( char *fbuf, int j, char *rplcstr, int len )
{
 int k=0;
 strcpy( word1, rplcstr );
 word1[len] = '\0';	/* Truncate to len. */
 prepad_with_whitespace( word1, len );
 while ((j+k < filesz) && (word1[k] != '\0')) 
  {
    fbuf[j+k] = word1[k];
    k++;
  }
}


void lookup_replacement( char *fbuf, int j, char *rplcstr1, char *rplcstr2 )
{ /* Coming into here, fbuf[j] will be pointing at a replicate of scrchstrng1. */
  /* Look at the chars direcly after that, which will be a line-label. */
  /* Look up the line-label value. */
  /* If found, then prepare replacement string(s) accordingly. */
  /* Otherwise, prepare white-space replacement strings. */
 struct nvpair *item;
 char label[100];
 int k, m1;

 k = strlen( srchstrng1 );
 if (replacement_type == 1)
  { /* The following 3 digits after marker will be the label index. */
    for (m1=0; m1 < 3; m1++) label[m1] = fbuf[j+k+m1];
    label[3] = '\0';
  }
 else
  { /* The string following the marker, terminated by '_' will be the label. */
    m1 = 0;
    while (fbuf[j+k+m1] != '_')
     {
      label[m1] = fbuf[j+k+m1];
      m1++;
      if (m1 > srch1_len - k)
	{ label[m1] = '\0';  printf("Error: Unterminated marker-label '%s'.\n", label );  exit(1); }
     }
   label[m1] = '\0';
  }
 /* Lookup the label-value. */
 item = results_list;
 while ((item != 0) && (strcmp( item->label, label ) != 0)) item = item->nxt;
 if (item != 0)
  {
   /* Now we have a value. Must prepare the replacement string(s), and to be of the exact lengths. */
   prepare_replacement_strings( item, rplcstr1, srch1_len, rplcstr2, srch2_len );
  }
 else
  {
   set_with_whitespaces( rplcstr1, srch1_len );
   set_with_whitespaces( rplcstr2, srch2_len );
  }
}


void lookup_line_label( char *label, char *rplcstr, int len )
{
 double value;
 struct nvpair *item;
 item = results_list;
 while ((item != 0) && (strcmp( item->label, label ) != 0)) item = item->nxt;
 if (item != 0)
  {
   if ((replacement_type == 1) && (sscanf(item->value,"%lf", &value) == 1))
    { /*isnumeric*/
      sprintf(rplcstr,"%d", m_round( value ) );
    } /*isnumeric*/
   else
    strcpy( rplcstr, item->value );
   if (verbose) printf("	Replacing with '%s', but truncating to length %d\n", rplcstr, len );
   //   rplcstr[ strlen(label) ] = '\0';
   rplcstr[ len ] = '\0';
   prepad_with_whitespace( rplcstr, len );
   // while (strlen( rplcstr ) < len) strcat( rplcstr, " " );
  }
 else
  set_with_whitespaces( rplcstr, len );  /* Not found, so replace with blank whitespace. */
 if (verbose) printf("Replacing '%s' with '%s\n", label, rplcstr );
}


void replace_pdf( char *fname )
{
 int j, ret, lastcentsindx=-1;
 struct stat fstatbuf;
 FILE *fptr;

 if (outputname == 0)
  {
   /* Create a new name for the output file.  "root_new.pdf" */ 
   outputname = (char *)malloc( strlen(results_file) + 10 );
   strcpy( outputname, results_file );
   j = strlen(outputname) - 1;
   while ((j >= 0) && (outputname[j] != '.')) j--;
   if (j < 0) j = strlen(outputname);
   outputname[j] = '\0';
   strcat( outputname, ".pdf" );
  }

 /* Determine the original file's size. */
 ret = stat( fname, &fstatbuf );
 if (ret != 0) { printf("Error seeing file '%s.\n", fname );  return; }
 filesz = fstatbuf.st_size;

 /* Allocate the working character buffer to hold the file's contents. */
 fbuf = (char *)malloc( filesz + 100 );
 if (fbuf == 0) { printf("Error: Out of memory. Not doing file '%s'.\n", fname );  return; }

 /* Read the file into the character-buffer. */
 fptr = fopen( fname, "rb" );
 if (fptr == 0) { printf("Error: Could not read file '%s'.\n", fname );  return; }
 j = 0;
 fbuf[j++] = getc(fptr);
 while (!feof(fptr))
  {
   if (j > filesz) { printf("Error: File '%s' running beyond expected length.\n", fname );  return; }
   fbuf[j++] = getc(fptr);
  }
 fclose(fptr);
 if (filesz != j -1) printf("Warning: FILE length: expected %d, but read actual %d\n", filesz, j );

 /* Do replacements here. */
 j = 0;
 while (j < filesz)
  { /*jloop*/
   if (chkstr( fbuf, j, srchstrng1 ))
    {
     lookup_replacement( fbuf, j, rplcstr1, rplcstr2 );
     replace_str( fbuf, j, rplcstr1 );
     if (lastcentsindx > 0)
      {
	replace_str( fbuf, lastcentsindx, rplcstr2 );
	lastcentsindx = -1;
      }
     else
     if ((replacement_type == 2) && (force_cents))
      {
	int n1;
	n1 = j;
	while ((n1 < filesz) && (!chkstr( fbuf, n1, srchstrng2 ))) n1++;
	if (n1 < filesz)
	 replace_str( fbuf, n1, rplcstr2 );
      }
    }
   if ((replacement_type == 2) && (srchstrng2 != 0) && (srchstrng2[0] != '\0') && (chkstr( fbuf, j, srchstrng2 )))
    {
     lastcentsindx = j;
    }
   if (alias_list)
    { /*Check all the aliases*/
      struct nvpair *item;
      item = alias_list;
      while ((item != 0) && (chkstr( fbuf, j, item->value ) == 0)) item = item->nxt;
      if (item != 0)
       {
	if (verbose) printf("Found '%s', looking up '%s'\n", item->value, item->label );
	lookup_line_label( item->label, word1, strlen(item->value) );
	if (verbose) printf("	REPLACING with: '%s'\n", word1 );
	replace_str( fbuf, j, word1 );
       }
    }
   if (symb_list)
    { /*Check all the direct symbol definitions for direct replacements.*/
      struct nvpair *item;
      item = symb_list;
      while ((item != 0) && (chkstr( fbuf, j, item->label ) == 0)) item = item->nxt;
      if (item != 0)
       {
	if (verbose) printf("Replacing symbol '%s' with '%s'\n", item->label, item->value );
	replace_str_direct( fbuf, j, item->value, strlen(item->label) );
       }
    }
   j++;
  } /*jloop*/

 /* Finally, replace any remaining markers with white-space. */
 set_with_whitespaces( rplcstr1, srch1_len );
 set_with_whitespaces( rplcstr2, srch2_len );
 j = 0;
 while (j < filesz)
  {
   if (chkstr( fbuf, j, srchstrng1 ))
    {
     replace_str( fbuf, j, rplcstr1 );
    }
   if ((srchstrng2[0] != '\0') && (chkstr( fbuf, j, srchstrng2 )))
    {
     replace_str( fbuf, j, rplcstr2 );
    }
   j++;
  } 

 /* Write the character-buffer out to the output file. */
 printf("Writing to:  %s\n", outputname );
 fptr = fopen( outputname, "wb" );
 if (fptr == 0) { printf("Error: Could not write file '%s'.\n", outputname );  return; }
 j = fwrite( fbuf, sizeof(char), filesz, fptr );
 fclose(fptr);
 if (j != filesz) printf("Error: Wrote only %d characters instead of %d to file %s\n", j, filesz, outputname );
}


/* --------------------- Main ----------------------------- */
int main( int argc, char *argv[] )
{
 int k=1, pp=1;
 char *marked_pdf_file=0;

 while (k < argc)
  {
   if (strcmp( argv[k], "-readSchedinput" ) == 0)
    {
	k++;
	if (k == argc) { printf("Missing filename after '%s'\n", argv[k-1] );  exit(1); }
	read_tax_input_for_schedule_data( argv[k] );
    }
   else
   if (strcmp( argv[k], "-o" ) == 0)
    {
	k++;
	if (k == argc) { printf("Missing filename after '%s'\n", argv[k-1] );  exit(1); }
	outputname = strdup( argv[k] );
    }
   else
   if (strcmp( argv[k], "-verbose" ) == 0)
    {
	verbose = 1;
    }
   else
   if (strcmp( argv[k], "-form8949_bnr" ) == 0)
    {
	form8949_bnr = 1;	/* Set to do "Basis Not Reported" version of 8949 form. */
    }
   else
    {
     switch (pp)
      {
       case 1:  read_metadata( argv[k] );
	 break;
       case 2:  strcpy( results_file, argv[k] );
		read_in_results_file( argv[k] );	     
 	 break;
       default:
	   if (marked_pdf_file)
	    { printf("Error: Too many files on %s command-line (%s).", argv[0], argv[k] );  exit(1); }
	   marked_pdf_file = strdup( argv[k] );
       }
     pp++;
    }
   k++;
  }

 if ((form_is != 0) && (strcmp(form_is,"Form_8949") == 0))
  Clear_8949_Entries( cg_list1, cg_list2 );

 if ((!detected_Sched_D) && (form_is != 0) && (strcmp(form_is,"Sched_D") == 0))
  Clear_Sched_D_Entries();

 if (marked_pdf_file == 0) { printf("Error: Missing pdf source file on %s command-line.", argv[0] );  exit(1); }
 replace_pdf( marked_pdf_file );
 return 0;
}
